/*
 * fsutil.c
 *
 * Functions for file system utility.
 *
 * Original working version written by:
 * Peter Brant <Peter.Brant@wicourts.gov>
 *
 * Adapted for submission to PostgreSQL community by:
 * Kevin Grittner <Kevin.Grittner@wicourts.gov>
 *
 * Copyright (c) 2006, Wisconsin Supreme Court
 * ALL RIGHTS RESERVED;
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHOR OR DISTRIBUTORS HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHOR AND DISTRIBUTORS HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 */

#include "fsutil.h"

PG_MODULE_MAGIC;

static void log_error(const char *);

static char **_fs_file_system(void);
static int64 _fs_get_total_kb(char *);
static int64 _fs_get_avail_kb(char *);


static char **_fs_file_system() {

#if defined(WIN32)

    TCHAR temp[BUFSIZE];  
    short driveCount = 0; 
    temp[0] = '\0';

    if (GetLogicalDriveStrings(BUFSIZE-1, temp)) {
        char *p = temp;
        char **result;
        char **resultp;

        do {
            driveCount++;
            while (*p++);
        } while (*p); 

        result = palloc(sizeof(char*)*(driveCount+1));
        if (result == NULL) {
            return NULL;
        }
        result[driveCount] = NULL;
        resultp = result;

        p = temp;
        do {
            int slen = strlen(p);
            *resultp = palloc(slen+1);
            if (result == NULL) {
                return NULL;
            }
            strcpy(*resultp, p);
            while (*p++);
            resultp++;
        } while (*p);

        return result;
    }
    else {
        log_error("GetLogicalDriveStrings()");

        return NULL;
    }
    
#else

    FILE *mounted;
    struct mntent *entry;
    char *strings[MAX_VOL_COUNT];
    char **temp;
    int fsCount;
    char **result;
    char **resultp;
    int i;

    mounted = fopen(_PATH_MOUNTED, "r");
    if (mounted == NULL) {
        log_error("fopen()");
        return NULL;
    }

    fsCount = 0;
    temp = strings;
    while ( (entry = getmntent(mounted)) != NULL) {
        char *mnt_dir;

        if (entry->mnt_dir == NULL 
                || (HAS_MNT_TYPE(entry, MNTTYPE_IGNORE) || HAS_MNT_TYPE(entry, MNTTYPE_SWAP))) {
            continue;
        }

        mnt_dir = pstrdup(entry->mnt_dir);
        if (mnt_dir == NULL) {
            endmntent(mounted);
            return NULL;
        }
        *temp = mnt_dir;
        temp++;
        fsCount++;
    }
    endmntent(mounted);

    result = palloc(sizeof(char *)*(fsCount+1));
    if (result == NULL) {
        return NULL;
    }
    result[fsCount] = NULL;

    temp = strings;
    resultp = result;
    for (i = 0; i < fsCount; i++) {
        *resultp++ = *temp++;
    }

    return result;

#endif

}

static int64 _fs_get_total_kb(char *s)
{

#if defined(WIN32)

    int64 total;
    BOOL ok;

    ok = GetDiskFreeSpaceEx(s, NULL, 
            (PULARGE_INTEGER)&total, NULL);

    if (ok) {
        return total / 1024;
    }
    else {
        return 0;
    }

#else

    int error;
    struct statvfs resp;

    error = statvfs(s, &resp);
    if (! error) {
        return (int64)resp.f_frsize*(int64)resp.f_blocks/1024;
    } 
    else {
        log_error("statvfs()");
        return -1;
    }

#endif

}

static int64 _fs_get_avail_kb(char *s)
{

#if defined(WIN32)

    int64 avail;
    BOOL ok;

    ok = GetDiskFreeSpaceEx(s, 
            (PULARGE_INTEGER)&avail, NULL, NULL);

    if (ok) {
        return avail / 1024;
    }
    else {
        return 0;
    }

#else

    int error;
    struct statvfs resp;

    error = statvfs(s, &resp);
    if (! error) {
        return (int64)resp.f_frsize*(int64)resp.f_bavail/1024;
    } 
    else {
        log_error("statvfs()");
        return -1;
    }

#endif

}

static void log_error(const char *lpszFunction)
{

#if defined(WIN32)

    /* 
     * From
     * http://msdn.microsoft.com/library/default.asp?url=/library/en-us/debug/base/retrieving_the_last_error_code.asp
     */
    LPVOID lpMsgBuf;
    LPVOID lpDisplayBuf;
    DWORD dw = GetLastError(); 

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    lpDisplayBuf = LocalAlloc(LMEM_ZEROINIT, 
        _tcslen(lpMsgBuf)+_tcslen(lpszFunction)+40); 
    wsprintf(lpDisplayBuf, 
      "%s failed with error %d: %s", 
      lpszFunction, dw, lpMsgBuf); 

    elog(ERROR, lpDisplayBuf);

    LocalFree(lpMsgBuf);
    LocalFree(lpDisplayBuf);

#else

    char buf[BUFSIZE];
    char *buf2;
    
    // Man page appears to be wrong here, work around as best we can
    buf[0] = '\0';
    buf2 = (char*)strerror_r(errno, buf, BUFSIZE);
    if (buf2 == NULL || (int)buf2 == -1) {
        buf2 = buf;
    }

    elog(ERROR, "%s failed with error %d: %s", lpszFunction, errno, buf2);

#endif

}


/*
 * SQL function: fs_get_total_kb(text) returns bigint
 * The parameter is a volume (such as a Windows drive letter or a Linux mount point)
 * as would be returned from the fs_file_system() function.
 */
PG_FUNCTION_INFO_V1(fs_get_total_kb);
Datum
fs_get_total_kb(PG_FUNCTION_ARGS)
{
    char       *arg;
    int64      result;
    arg = _textout(PG_GETARG_TEXT_P(0));
    result = _fs_get_total_kb(arg);
    PG_RETURN_INT64(result);
}

/*
 * SQL function: fs_get_avail_kb(text) returns bigint
 * The parameter is a volume (such as a Windows drive letter or a Linux mount point)
 * as would be returned from the fs_file_system() function.
 */
PG_FUNCTION_INFO_V1(fs_get_avail_kb);
Datum
fs_get_avail_kb(PG_FUNCTION_ARGS)
{
    char       *arg;
    int64      result;
    arg = _textout(PG_GETARG_TEXT_P(0));
    result = _fs_get_avail_kb(arg);
    PG_RETURN_INT64(result);
}

/*
 * SQL function: fs_file_system() returns a result set with one text column.
 * The text contains the volume, such as a Windows drive letter or a Linux mount point.
 */
PG_FUNCTION_INFO_V1(fs_file_system);
Datum
fs_file_system(PG_FUNCTION_ARGS)
{
    FuncCallContext *funcctx;
    MemoryContext oldcontext;
    char **data;

    int call_cntr;
    int max_calls;

    if (SRF_IS_FIRSTCALL()) {
        char **temp;
        int fsCount = 0;

        funcctx = SRF_FIRSTCALL_INIT();

        oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
        
        data = _fs_file_system();

        fsCount = 0;
        if (data != NULL) {
            temp = data;
            while (*temp++) {
                fsCount++;
            }
        }

        funcctx->max_calls = fsCount;
        funcctx->user_fctx = data;

        MemoryContextSwitchTo(oldcontext);
    }

    funcctx = SRF_PERCALL_SETUP();
    call_cntr = funcctx->call_cntr;
    max_calls = funcctx->max_calls;

    data = funcctx->user_fctx;

    if (call_cntr < max_calls) {
        SRF_RETURN_NEXT(funcctx, PointerGetDatum(_textin(data[call_cntr])));
    } 
    else {
        SRF_RETURN_DONE(funcctx);
    }
}
